#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_defender_kick.h"
#include "bhv_basic_offensive_kick.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "body_kubo_dribble.h"
#include "body_pass_clone.h"
#include "body_safety_pass.h"
#include "strategy.h"
#include "body_clear_clone.h"
#include "offense_system.h"

#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_dribble.h>

#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

bool
Bhv_DefenderKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderKick"
                        ,__FILE__, __LINE__ );
    
   const rcsc::WorldModel & wm = agent->world();

    //pass near front mate
    if( wm.opponentsFromSelf().empty() )return false;
    double nearest_opp_dist = (wm.opponentsFromSelf().front()->pos()
                               - wm.self().pos()).r();
    const rcsc::PlayerObject nearest_mate = *wm.teammatesFromSelf().front();
    double nearest_mate_dist = (nearest_mate.pos() - wm.self().pos()).r();
    
    if( wm.self().pos().x >= -35 && wm.self().pos().x <= 35
        && nearest_mate.pos().x >= wm.self().pos().x
        && nearest_opp_dist >= 3 && nearest_mate_dist <= 3)
    {
        if(rcsc::Body_KickMultiStep(nearest_mate.pos(),1.5,
                                    false).execute(agent)){
            agent->debugClient().addMessage( "Choi" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: choi pass"
                                ,__FILE__, __LINE__ );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return true;
        }
    }
    
    static int S_hold_count = 0;

    //pass(no stamina)
    rcsc::Vector2D target;
    if( wm.self().stamina() < 2500 &&
        Body_PassClone::get_best_pass(wm,&target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( target ) ==
            Strategy::BA_ShootChance ){
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: no stamina pass"
                                ,__FILE__, __LINE__ );
            Body_PassClone().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }
    
    //kuboDribble        
    if( S_hold_count >= 3 && Body_KuboDribble(KUBO,0,2,3).execute(agent)){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //safety pass
    if( S_hold_count >= 5 && Body_SafetyPass().execute(agent)){
            
        agent->debugClient().addMessage( "safetyPass" );

        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }
            
    //kuboDribble    
    if( Body_KuboDribble(KUBO,0,3,3).execute(agent) ){
        agent->setNeckAction( new rcsc::Neck_ScanField() );
        S_hold_count = 0;
        return true;
    }

    //pass
    if( Body_PassClone::get_best_pass(wm,&target,NULL,NULL) ){
        //ŨǤϥѥʤ
        if( Strategy::get_ball_area( wm.ball().pos() ) !=
            Strategy::BA_ShootChance ||
            Strategy::get_ball_area( target ) ==
            Strategy::BA_ShootChance ){
            Body_PassClone().execute(agent);
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //clear
    if( Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_Danger ){
        if( Body_ClearClone().execute(agent) ){
            agent->setNeckAction( new rcsc::Neck_ScanField() );
            S_hold_count = 0;
            return true;
        }
    }

    //avoid near opp
    rcsc::Vector2D min_cong_point = OffenseSystem::getMinCongestionPoint(agent,
                                                                         1.5);
    if( min_cong_point.absX() < 47.0 && min_cong_point.absY() < 30.0 ){
        rcsc::Body_Dribble( min_cong_point,
                            0.5,
                            rcsc::ServerParam::i().maxPower(),
                            2 ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        ++S_hold_count;
        return true;
    }

    //hold ball
    rcsc::Body_HoldBall().execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
    ++S_hold_count;


    return true;
}
