#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_defender_move.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_turn_to_ball.h>
#include <rcsc/action/body_turn_to_point.h>
#include "bhv_basic_tackle.h"

#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "body_zulu.h"
#include "strategy.h"


/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_DefenderMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    
    ////////////zulu////////////
    if( Body_Zulu().hear(agent) ){
               agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
               return true;
    }

    // tackle
    if ( Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }

    // intercept
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    if ( ! wm.existKickableTeammate()
         && ( self_min <= 3
              || ( self_min < mate_min + 2
                   && self_min < opp_min + 3 )
              )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept self_min=%d mate_min=%d opp_min=%d"
                            ,__FILE__, __LINE__,
                            self_min,mate_min,opp_min);
        rcsc::Body_Intercept().execute( agent );
        
        if ( wm.ball().distFromSelf()
             < rcsc::ServerParam::i().visibleDistance() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }


    const double dash_power = getDashPower( agent, M_home_pos );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    // defense line
//     rcsc::Vector2D target_point= M_home_pos;
//     if( wm.self().pos().x > M_home_pos.x
//         && wm.ball().pos().x > wm.self().pos().x + 3){

//         double tmpx = std::floor( wm.ball().pos().x * 0.1 ) * 10.0 - 3.0;
//         if( tmpx > 0 )tmpx = 0;
//         if( tmpx > M_home_pos.x + 5.0 )tmpx = M_home_pos.x + 5.0;
//         target_point.x = tmpx;
//         if( wm.defenseLineX() < target_point.x - 1.0 ){
//             target_point.x = wm.defenseLineX() + 1.0;
//         }
//         target_point.y = M_home_pos.y;
//         double home_y_diff = wm.ball().pos().y - M_home_pos.y;
//         target_point.y += home_y_diff * 0.1;

//         //롼ѥк˥ǥե󥹥饤󲼤
//         if( Strategy::get_ball_area( target_point ) != Strategy::BA_Danger ){
//             if( target_point.x <= -20 )
//                 target_point.x -= 5;
//             if( target_point.x < 0 && target_point.x > -20 )
//                 target_point.x += 5 * (target_point.x)/20.0;
//         }

//         agent->debugClient().setTarget( target_point );
        
//         if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power
//                  ).execute( agent ) )
//         {
//             rcsc::Vector2D face_point( wm.self().pos().x , 0.0 );
//             face_point.y = ( wm.ball().pos().y < wm.self().pos().y ? -40.0 : 40.0 );
//             rcsc::Body_TurnToPoint( face_point ).execute( agent );
//         }
        
//         if ( wm.existKickableOpponent()
//              && wm.ball().distFromSelf() < 18.0 )
//         {
//             agent->setNeckAction( new rcsc::Neck_TurnToBall() );
//         }
//         else
//         {
//             agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
//         }
        
//         return true;
        
//     }
    //------------------------------//


    //mark opponent
//     if( Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_Danger ){
//         const rcsc::PlayerObject * mark_opp = getMarkTarget( agent,
//                                                              M_home_pos );
//         if( mark_opp != NULL ){
//             rcsc::AngleDeg block_angle = (wm.ball().pos() - mark_opp->pos()).th();
//             rcsc::Vector2D block_point
//                 = mark_opp->pos() + rcsc::Vector2D::polar2vector(0.2,block_angle);
//             block_point.x -= 0.1;
//             if(block_point.x < wm.self().pos().x - 1.5){
//                 block_point.y = wm.self().pos().y;
//             }
//             agent->debugClient().addMessage( " MarkMove%.0f", dash_power );
//             agent->debugClient().setTarget( block_point );
            
//             if ( ! rcsc::Body_GoToPoint( block_point, dist_thr, dash_power
//                      ).execute( agent ) )
//             {
//                 rcsc::Body_TurnToBall().execute( agent );
//             }
            
//             if ( wm.existKickableOpponent()
//                  && wm.ball().distFromSelf() < 18.0 )
//             {
//                 agent->setNeckAction( new rcsc::Neck_TurnToBall() );
//             }
//             else
//             {
//                 agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
//             }
//         }
//     }

    //block our goal
    if( wm.self().pos().x < -43.0
        && (wm.self().unum() == 4 || wm.self().unum() == 5) ){
        bool block_flg = false;
        rcsc::Vector2D block_point( -49.0,
                                    rcsc::ServerParam::i().goalHalfWidth() - 3.0 );
        if( wm.ball().pos().y * M_home_pos.y < 0
            && wm.existKickableOpponent()
            && wm.ball().pos().x < -40
            && 5 < wm.ball().pos().absY() ){
            block_point.y = 0.5;
            block_flg = true;
        }
        if( wm.ball().pos().y * M_home_pos.y < 0
            && wm.existKickableOpponent()
            && wm.ball().pos().x < -40
            && 6 < wm.ball().pos().absY()
            && wm.ball().pos().absY() < 12.5 ){
            block_point.y = rcsc::ServerParam::i().goalHalfWidth() - 1.5;
            block_flg = true;
        }
        if( M_home_pos.y < 0 ) block_point.y *= -1.0;

        if(block_flg){
            agent->debugClient().addMessage( "block" );
            agent->debugClient().setTarget( block_point );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: block our goal"
                                ,__FILE__, __LINE__ );
            
            if ( ! rcsc::Body_GoToPoint( block_point, dist_thr, dash_power
                     ).execute( agent ) )
            {
                rcsc::Body_TurnToBall().execute( agent );
            }
            
            if ( wm.existKickableOpponent()
                 && wm.ball().distFromSelf() < 18.0 )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            }
            
            return true;
        }
    }
    
    //normal move
    rcsc::Vector2D new_home_pos = M_home_pos;
        
    //nearest_DFõ
    int nearest_DF = 0;
    if( ! wm.existKickableTeammate() )
    {
        double min_dist = 100, tmp_dist;
        // find DF nearest to ball
        for(int i=2;i<=5;i++){
            if(  wm.teammate(i) != static_cast< rcsc::PlayerObject * >( 0 ) ){
                tmp_dist = wm.teammate(i)->pos().dist( wm.ball().pos() );
                if( tmp_dist < min_dist
                    && wm.teammate(i)->pos().x < wm.ball().pos().x + 1 ){
                    min_dist = tmp_dist;
                    nearest_DF = i;
                }
            }
        }
        
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: nearestDF = %d"
                            ,__FILE__, __LINE__, nearest_DF );

        //--롼ѥк˥ǥե󥹥饤󲼤
        if( nearest_DF != wm.self().unum()
            && Strategy::get_ball_area( M_home_pos ) != Strategy::BA_Danger )
        {
            if( M_home_pos.x < 0 && M_home_pos.x > -20 )
                new_home_pos.x -= -5 * (new_home_pos.x)/20.0;
            if( M_home_pos.x <= -20 )
                new_home_pos.x -= 5;
        }

        //--ɥ֥˸֤˹Ԥ
        if( nearest_DF == wm.self().unum()
            && Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_Danger
            && wm.ball().pos().x < wm.self().pos().x + 7
            && wm.ball().pos().x < wm.defenseLineX() + 10)
        {
            rcsc::AngleDeg block_angle
                = (rcsc::Vector2D(-60,0) - wm.ball().pos()).th();
            
            if( wm.self().pos().absY() < 6 ){
                new_home_pos
                    = wm.ball().pos()
                    + rcsc::Vector2D(-5.0,0);
            } else if( wm.self().pos().x < -45 ){
                if( wm.self().pos().y >= 0 ){
                    new_home_pos
                        = wm.ball().pos()
                        + rcsc::Vector2D(0,-3.0);
                } else {
                    new_home_pos
                        = wm.ball().pos()
                        + rcsc::Vector2D(0,3.0);
                }
            } else {
                new_home_pos
                    = wm.ball().pos()
                    + rcsc::Vector2D::polar2vector(5.0,block_angle);
            }
            agent->debugClient().addMessage( "blockD" );
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: block dribble"
                                ,__FILE__, __LINE__);
        }
    }
        
    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( new_home_pos );
    
    if ( ! rcsc::Body_GoToPoint( new_home_pos, dist_thr, dash_power
             ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }

    return true;
    
}


double
Bhv_DefenderMove::getDashPower( const rcsc::PlayerAgent * agent,
              const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/

    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();


    /*------defender-----*/
    if( wm.ball().pos().x < wm.self().pos().x-0.5 )
        return rcsc::ServerParam::i().maxPower();
    
    if( wm.ball().pos().x >= -35.0 && wm.existKickableTeammate()
        && wm.self().stamina() <= 3500 )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;
        return dash_power;
    }
    
    else if( wm.ball().pos().x <= -15.0 ){
        if( wm.ball().pos().x <= wm.defenseLineX() + 10 )
            return rcsc::ServerParam::i().maxPower();
        else
            return std::min( my_inc * 1.7,
                             rcsc::ServerParam::i().maxPower() );
    }
    
    else if(wm.ball().pos().x >= 0.0){
        if(wm.self().stamina() <= 3500){
            //rest
            return std::max( 0.0, my_inc - 25.0 );
        } else {
            dash_power = std::min( my_inc * 1.5,
                                   rcsc::ServerParam::i().maxPower() );
            return dash_power;
        }
        
    } else {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        return dash_power;
    }
    /*-------------------*/


    dash_power = 100.0;

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: correct DF line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable teammate. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    } 
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: normal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    return dash_power;

}


const rcsc::PlayerObject *
Bhv_DefenderMove::getMarkTarget( rcsc::PlayerAgent * agent,
                                 const rcsc::Vector2D & home_pos )
{

    const rcsc::WorldModel & wm = agent->world();
    
    double dist = 1000;
    const rcsc::PlayerObject * opp
        = wm.getOpponentNearestTo( home_pos, 1, &dist );

    if( !opp || (wm.existKickableTeammate()
                 && opp->distFromBall() > 2.5 ) ){
        return NULL;
    }

    if( dist > 7 && home_pos.x < opp->pos().x ){
        return NULL;
    }

    const rcsc::PlayerPtrCont::const_iterator
        end = wm.teammatesFromSelf().end();
    for( rcsc::PlayerPtrCont::const_iterator
             it = wm.teammatesFromSelf().begin();
         it != end; ++it)
    {
        if( (*it)->pos().dist( opp->pos() ) < dist ){
            return NULL;
        }
    }

    return opp;

}
