#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_forward_move.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_turn_to_ball.h>
#include "bhv_basic_tackle.h"

#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "body_zulu.h"
#include "offense_system.h"
#include "strategy.h"

#include <rcsc/action/intention_time_limit_action.h>

#include <iostream>
using namespace std;


/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_ForwardMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_ForwardMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    
    //zulu
    if( Body_Zulu().hear(agent) ){
               agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
               return true;
    }
    
    // tackle
    if ( wm.self().pos().x < 10
         && Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }

    // intercept
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ! wm.existKickableTeammate()  //magic number
         && ( self_min <= 3
              || ( self_min < mate_min + 3
                   && self_min < opp_min + 4 )
              )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: intercept"
                            ,__FILE__, __LINE__ );
        rcsc::Body_Intercept().execute( agent );
        if ( wm.ball().distFromSelf()
             < rcsc::ServerParam::i().visibleDistance() )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );;
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return true;
    }

    //üäǥѥäƤɤ
    static int S_repeat_count = 0;
    if( !wm.existKickableTeammate()
        && abs(wm.ball().pos().y) >= 20 && wm.ball().pos().x <= -25 )
    {
        ++S_repeat_count;
        if( (wm.ball().pos().y <= 0 && wm.self().unum() == 9 ||
             wm.ball().pos().y >= 0 && wm.self().unum() == 10 ) &&
            S_repeat_count >= 25) //magic number 
        {
            //intercept
            agent->setIntention(
                new rcsc::IntentionTimeLimitAction( new rcsc::Body_Intercept(),
                                                    new rcsc::Neck_TurnToBallOrScan(),
                                                    NULL,
                                                    15,
                                                    wm.time() ) );
            rcsc::Body_Intercept().execute(agent);
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return true;
        }
    } else {
        S_repeat_count = 0;
    }
    
    const double dash_power = getDashPower( agent, M_home_pos );
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;
    rcsc::Vector2D new_home_pos = M_home_pos;
    rcsc::Vector2D center_home_pos = M_home_pos;

    //ɤ˹ꤹʤ褦
    const double half_width = 15; //magic number
    if( wm.teammate(11) != static_cast< rcsc::PlayerObject * >( 0 )
        && (wm.self().unum() == 9 || wm.self().unum() == 10) ){
        if( center_home_pos.y > 0
            && wm.teammate(11)->pos().y + half_width < wm.self().pos().y){
            center_home_pos.y = wm.teammate(11)->pos().y + half_width;
        }
        if( center_home_pos.y <= 0
            && wm.teammate(11)->pos().y - half_width > wm.self().pos().y){
            center_home_pos.y = wm.teammate(11)->pos().y - half_width;
        }
    }
    
    //zuluѤhome_posĴ
    double offside = wm.offsideLineX();
    const double GO = -15;
    const double ARRIVE = 5;
    if( wm.ball().pos().x > GO && wm.ball().pos().x < ARRIVE ){
        new_home_pos.x +=  (offside - new_home_pos.x)
            * (new_home_pos.x - GO)
            / (ARRIVE - GO);
        new_home_pos.y = OffenseSystem::getLineCongestion(agent,
                                                          new_home_pos.x,
                                                          center_home_pos);
        double mates_dist=0;
        wm.getTeammateNearestTo(new_home_pos,5,&mates_dist);
        if( new_home_pos.absY() >= 27 )
            new_home_pos.y = center_home_pos.y;
        if( wm.self().pos().dist(new_home_pos) > mates_dist )
            new_home_pos.y = center_home_pos.y;
    }
    if( wm.ball().pos().x >= ARRIVE ){
        new_home_pos.x = offside;
        new_home_pos.y = OffenseSystem::getLineCongestion(agent,
                                                          new_home_pos.x,
                                                          center_home_pos);
        double mates_dist=0;
        wm.getTeammateNearestTo(new_home_pos,5,&mates_dist);
        if( new_home_pos.absY() >= 27 )
            new_home_pos.y = center_home_pos.y;
        if( wm.self().pos().dist(new_home_pos) > mates_dist )
            new_home_pos.y = center_home_pos.y;
    }
    if( wm.ball().pos().x > 32 )
        new_home_pos.y = center_home_pos.y;

    //shoot chance move
    if(Strategy::get_ball_area( wm.ball().pos() ) == Strategy::BA_ShootChance)
    {
        rcsc::Vector2D target_point = M_home_pos;
        if( wm.ball().pos().y * wm.self().pos().y < 0
            && wm.ball().pos().x > 40
            && 6 < wm.ball().pos().absY()
            && wm.ball().pos().absY() < 17 ){
            rcsc::Rect2D goal_area( rcsc::Vector2D(52.5 - 8.0, -6.0),
                                    rcsc::Vector2D(52.5,6.0) );
            if( !wm.existTeammateIn( goal_area,10,true ) ){
                target_point.x = wm.ball().pos().x + 2.0;
                if( target_point.x > wm.offsideLineX() - 1.0 ){
                    target_point.x = wm.offsideLineX() - 1.0;
                }
                if( target_point.x > 49.0 ) target_point.x = 49.0;
                target_point.y = -1.0;
                if( wm.ball().pos().y > 0.0 ) target_point.y *= -1.0;
            }
        }
        agent->debugClient().addMessage( "shootChanceMove" );
        agent->debugClient().setTarget( target_point );
        
        if ( ! rcsc::Body_GoToPoint( target_point, 0.5,
                                     rcsc::ServerParam::i().maxPower()
                 ).execute( agent ) )
        {
            rcsc::Body_TurnToBall().execute( agent );
        }
        
        if ( wm.existKickableOpponent()
             && wm.ball().distFromSelf() < 18.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        
        return true;
    }
    
    //normal move

    //--꤬֤ʤ褦
    if( Strategy::get_ball_area( wm.ball().pos() ) != Strategy::BA_Danger
        && wm.existKickableTeammate()
        && M_home_pos.dist( wm.ball().pos() ) <= 2 )
    {
        if( wm.self().pos().x > wm.ball().pos().x )
            new_home_pos.x += 3.0;
        else
            new_home_pos.x -= 3.0;
    }
    
    agent->debugClient().addMessage( " BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( new_home_pos );

    if ( ! rcsc::Body_GoToPoint( new_home_pos, dist_thr, dash_power
             ).execute( agent ) )
    {
        rcsc::Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
    
        return true;
}



double
Bhv_ForwardMove::getDashPower( const rcsc::PlayerAgent * agent,
              const rcsc::Vector2D & target_point )
{
    static bool s_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // check recover

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/

    double dash_power = 100.0;
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();


    /*---FW---*/
    if ( wm.self().pos().x > wm.offsideLineX() + 0.5 )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
        return dash_power;
    }
    
    if( wm.ball().pos().x <= -15.0 ){
        if(wm.self().stamina() <= 3000){
            //rest
            return std::max( 0.0, my_inc - 25.0 );
        } else {
            dash_power = std::min( my_inc * 1.5,
                                   rcsc::ServerParam::i().maxPower() );
            return dash_power;
        }
    } else {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        return dash_power;
    }
    
    /*-------------------*/


    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: correct DF line. keep max power"
                            ,__FILE__, __LINE__ );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
    }
// exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: exist kickable teammate. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
// in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: in offside area. dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
// normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: normal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    return dash_power;

}

