#include "body_kubo_dribble.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_dribble.h>
#include <rcsc/geom/polygon_2d.h>

#include "offense_system.h"
#include <rcsc/action/intention_time_limit_action.h>
#include <rcsc/action/neck_scan_field.h>

#include <iostream>
using namespace std;


bool
Body_KuboDribble::execute(rcsc::PlayerAgent * agent)
{
    switch( M_type ){
    case KIWAMU:
        return kiwamuDribble(agent);
        break;
    case ONE:
        return oneOnOneDribble(agent);
        break;
    case KUBO:
        return kuboDribble(agent);
    default:
        break;
        
    }

    return true;
}

//M_rangeĹζ8äơŨʤ˥ɥ֥
bool
Body_KuboDribble::kiwamuDribble(rcsc::PlayerAgent * agent)
{

    const rcsc::WorldModel & wm = agent->world();

    double afraid_range = M_range;


    double rect_h_width = 3; //magic number
    
    
   //make rectangle vectors
    vector< rcsc::Vector2D > rect_vectors;
    rect_vectors.push_back( rcsc::Vector2D( 0,-rect_h_width ));
    rect_vectors.push_back( rcsc::Vector2D( 0, rect_h_width ));
    rect_vectors.push_back( rcsc::Vector2D( afraid_range,-rect_h_width ));
    rect_vectors.push_back( rcsc::Vector2D( afraid_range, rect_h_width ));

    //set kiwamu_point
    //0Ȥƻײ45٤7ޤǷ8
    rcsc::Polygon2D kiwamu_point[8];
    vector< rcsc::Vector2D > rect_points;
    for(int i=0;i<8;i++){

        for(unsigned int j=0;j<rect_vectors.size();j++)
            rect_points.push_back( wm.self().pos() + rect_vectors[j] );
        
        kiwamu_point[i].assign( rect_points );

        //rotate rectangle points
        for(unsigned int j=0;j<rect_vectors.size();j++)
            rect_vectors[j].rotate( 45.0 );

        rect_points.clear();

    }

    
    rcsc::Vector2D dribble_point =wm.self().pos();
    //kiwamu_point loop
    //õϤ뤿ѤˤʤäƤ
    for(int i=0;i<=4;i++)
    {
        int dir = i;
        if( dir != 0 && dir != 4 && wm.self().pos().y > 0 )
            dir = 8 - dir;
        
        if( !wm.existOpponentIn(kiwamu_point[dir], 10, true ) )
        {
            dribble_point = kiwamu_point[i].xyCenter();

            //add rectangle
            vector< rcsc::Vector2D > p = kiwamu_point[i].vertex();
            agent->debugClient().addLine( p[0], p[1] );
            agent->debugClient().addLine( p[1], p[3] );
            agent->debugClient().addLine( p[3], p[2] );
            agent->debugClient().addLine( p[2], p[0] );
            

            if(dribble_point.x>(rcsc::ServerParam::i().pitchHalfLength()-3))
                dribble_point.x=rcsc::ServerParam::i().pitchHalfLength()-3;
            if(dribble_point.y<(-rcsc::ServerParam::i().pitchHalfWidth()+3))
                dribble_point.y=-rcsc::ServerParam::i().pitchHalfWidth()+3;
            if(dribble_point.y>(rcsc::ServerParam::i().pitchHalfWidth()-3))
                dribble_point.y=rcsc::ServerParam::i().pitchHalfWidth()-3;
            
            rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                                "%s:%d: !kiwamu_dribble success!"
                                ,__FILE__, __LINE__ );
            
            rcsc::Body_Dribble(dribble_point,1.0,
                               rcsc::ServerParam::i().maxPower(),
                               5,false).execute(agent);
            return true;
        }

        //no reverse dir
        if( dir == 0 || dir == 4 ) continue;

        //reverse
        dir = 8 - dir;
        
        if( !wm.existOpponentIn(kiwamu_point[dir], 10, true ) )
        {
            dribble_point = kiwamu_point[i].xyCenter();

            if(dribble_point.x>(rcsc::ServerParam::i().pitchHalfLength()-3))
                dribble_point.x=rcsc::ServerParam::i().pitchHalfLength()-3;
            if(dribble_point.y<(-rcsc::ServerParam::i().pitchHalfWidth()+3))
                dribble_point.y=-rcsc::ServerParam::i().pitchHalfWidth()+3;
            if(dribble_point.y>(rcsc::ServerParam::i().pitchHalfWidth()-3))
                dribble_point.y=rcsc::ServerParam::i().pitchHalfWidth()-3;
            
            rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                                "%s:%d: !kiwamu_dribble success!"
                                ,__FILE__, __LINE__ );
            
            rcsc::Body_Dribble(dribble_point,1.0,
                               rcsc::ServerParam::i().maxPower(),
                               5,false).execute(agent);
            return true;
        }
        
    
    }
    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                        "%s:%d: !kiwamu_dribble failed!"
                        ,__FILE__, __LINE__ );
    return false;

}


bool
Body_KuboDribble::oneOnOneDribble(rcsc::PlayerAgent * agent){
    
    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                        "%s:%d: !one on one start!"
                        ,__FILE__, __LINE__ );
    
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();

    if( wm.self().pos().x < 0.0)return false;

    //get goalie info;
    double goalie_pos_x = 46.0;
    const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
		  it != opps_end;
		  ++it )
    {
		if( (*it)->goalie() )
        {
            goalie_pos_x = (*it)->pos().x - 5.0;
            break;
        }
    }

    //check cycle
    int self_cycle,opp_cycle,max_diff=-100;
    int best_idx=0;
    rcsc::Vector2D shoot_pos_candidate[5];
    for(int i=0;i<5;i++){

        if( wm.self().pos().y > 0 )
            shoot_pos_candidate[i].assign( goalie_pos_x, -6.0+(3.0*(double)i) );
        else
            shoot_pos_candidate[i].assign( goalie_pos_x, 6.0-(3.0*(double)i) );


        self_cycle = OffenseSystem::getPointReachCycle( agent,
                                                 shoot_pos_candidate[i],
                                                 wm.self());
        
        OffenseSystem::pointPredict( agent, shoot_pos_candidate[i],10,
                                     OPP, &opp_cycle, NULL,true);
        
        if( opp_cycle - self_cycle >= max_diff ){
            max_diff = opp_cycle - self_cycle;
            best_idx = i;
        }
    }
    rcsc::Vector2D shoot_pos = shoot_pos_candidate[best_idx];

    //checkOneOnOne
    bool free = true;
    bool goalie;
    vector< rcsc::Vector2D > points;
    points.push_back( rcsc::Vector2D( 52.0, 9.0));
    points.push_back( rcsc::Vector2D( 52.0, -9.0) );
    points.push_back( wm.self().pos() + rcsc::Vector2D(0,-0.1) );
    points.push_back( wm.self().pos() + rcsc::Vector2D(0,0.1) );
    const rcsc::Polygon2D poly( points );
    //opp loop
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
		  it != opps_end;
		  ++it )
    {
        //goalieֹƤʤѽ
        if( wm.self().pos().x <= 40 && abs((*it)->pos().y) <= 9
            && (*it)->pos().x >= 47 ){
            goalie = true;
        } else {
            goalie = false;
        }
        
		if( !( (*it)->goalie() || goalie ) && poly.contains( (*it)->pos() ) )
        {
            free = false;
            break;
        }
    }
    //mate loop
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromSelf();
    const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = mates.begin();
		  it != mates_end;
		  ++it )
    {
        if( poly.contains( (*it)->pos() ) )
        {
            free = false;
            break;
        }
    }

    //dribble
    if(free){
        if( (wm.self().pos()-shoot_pos).r() <= 1.0 ){
            rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                                "%s:%d: !One on One already reached!"
                                ,__FILE__, __LINE__,wm.self().body().abs() );
            return false;
        }
            

        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !One on One body=%f!"
                            ,__FILE__, __LINE__,wm.self().body().abs() );

        agent->debugClient().addLine( points[1],points[2] );
        agent->debugClient().addLine( points[2],points[3] );
        agent->debugClient().addLine( points[3],points[0] );
        agent->debugClient().addCircle( shoot_pos,2 );

        int dash_count=5;
        if( wm.self().pos().x > 41 )dash_count = 2;
        
//         static rcsc::Vector2D S_old_shoot_pos;
//         static int S_old_flg=false;
//         static int S_old_count=0;
        
//         if( wm.self().pos().x > wm.offsideLineX() - 5
//             && wm.self().pos().x < 36.0){
//             rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
//                                 "%s:%d: !One on One set Dribble Intention!"
//                                 ,__FILE__, __LINE__ );
//             S_old_flg = true;
//             S_old_shoot_pos = shoot_pos;
//         }

//         if( S_old_flg ){
//             S_old_count++;
//             rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
//                                 "%s:%d: !One on One Dribble Intention!"
//                                 ,__FILE__, __LINE__ );
//             rcsc::Body_Dribble( S_old_shoot_pos,
//                                 1.0,
//                                 rcsc::ServerParam::i().maxPower(),
//                                 dash_count,
//                                 false
//                 ).execute( agent );
//             if( S_old_count > 10 ){
//                 S_old_count = 0;
//                 S_old_flg = false;
//             }
//             return true;
//         }
        
        rcsc::Body_Dribble( shoot_pos,
                            1.0,
                            rcsc::ServerParam::i().maxPower(),
                            dash_count,
                            false
            ).execute( agent );
        return true;
        
    }

    return false;
    
}


//ȾM_distȾߤ˱äãͽ¬
//M_afraid;͵äɥ֥
bool
Body_KuboDribble::kuboDribble(rcsc::PlayerAgent * agent)
{
    double dist = M_dist;
    int kubo_afraid = M_afraid;
    
    rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                        "%s:%d: !start kubodribble!"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    bool flg=true;
    vector< rcsc::Vector2D > kubo_point;

    //dir from body to goal
    rcsc::AngleDeg want_dir = ( rcsc::Vector2D(50.0,0.0) - wm.self().pos() ).th();
    //set kubo_point
    rcsc::Vector2D tmp(1.0,0.0);
    tmp.setLength(dist);
    tmp.setDir( rcsc::AngleDeg( want_dir - 90));

    for(int i=0;i<13;i++){
        kubo_point.push_back( wm.self().pos() + tmp );

        if(kubo_point[i].x>(rcsc::ServerParam::i().pitchHalfLength()-2))
            kubo_point[i].x=rcsc::ServerParam::i().pitchHalfLength()-2;
        if(kubo_point[i].y<(-rcsc::ServerParam::i().pitchHalfWidth()+2))
            kubo_point[i].y=-rcsc::ServerParam::i().pitchHalfWidth()+2;
        if(kubo_point[i].y>(rcsc::ServerParam::i().pitchHalfWidth()-2))
            kubo_point[i].y=rcsc::ServerParam::i().pitchHalfWidth()-2;

        tmp.rotate(15);
    }

    //search safety direction
    int direction;
    flg = searchSafetyDir( agent,kubo_point,direction,kubo_afraid);

    //dribble or holdBall
    if(flg){
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !dribble to %d!"
                            ,__FILE__, __LINE__, direction );
        agent->debugClient().addLine
            ( wm.self().pos(),kubo_point[direction]);
	 
        rcsc::Body_Dribble(kubo_point[direction],
                           0.5,
                           rcsc::ServerParam::i().maxPower(),
                           1,
                           false).execute(agent);
    } else {
        rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
                            "%s:%d: !can't dribble!"
                            ,__FILE__, __LINE__ );
    }  

    return flg;

}

bool
Body_KuboDribble::searchSafetyDir(rcsc::PlayerAgent * agent,
                                  const vector< rcsc::Vector2D > & kubo_point,
                                  int &direction,
                                  const int kubo_afraid)
{
  int add,goal_dir=6;
  const rcsc::WorldModel & wm = agent->world();

  //no opp
  if( agent->world().opponents().empty() ){
	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !opponent is empty!"
								,__FILE__, __LINE__ );
	
	 direction = goal_dir;
	 return true;
  }

  int self_cycle;
  int opp_cycle;
  bool flg;
  
  //direction loop
  for(int i=0;i<=6;i++){

	 add = i;

	 //---check one side---	
	 flg = true;
	 if( wm.self().pos().y > 0.0 && add != 0)add *= -1;
	 direction = goal_dir + add;

	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !search dir= %d!"
								,__FILE__, __LINE__, direction );

	 self_cycle = OffenseSystem::getPointReachCycle(agent,
                                             kubo_point[direction],
                                             wm.self());
	 opp_cycle = OffenseSystem::pointPredict(agent,kubo_point[direction],
                                      10,OPP,NULL,NULL );

	 if(self_cycle != -1 ||
		 opp_cycle != -1 || 
		 opp_cycle - self_cycle < kubo_afraid){
		flg = false;
		rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								  "%s:%d: !afraid diff= %d, afraid= %d!"
								  ,__FILE__, __LINE__,
								  opp_cycle-self_cycle, kubo_afraid);
	 }
	 if(flg)break;
	

	 //---skip if direction==goal_dir---
	 if(add == 0)continue;

	
	 //---check other side---
	 add *= -1;
	 direction = goal_dir + add;
	 flg = true;

	 rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								"%s:%d: !search dir= %d!"
								,__FILE__, __LINE__, direction );

	 self_cycle = OffenseSystem::getPointReachCycle(agent,
                                             kubo_point[direction],
                                             wm.self());
     OffenseSystem::pointPredict(agent,kubo_point[direction],
                          10,OPP,&opp_cycle,NULL );

	 if(self_cycle <= 0 || 
		 opp_cycle <= 0 ||
		 opp_cycle - self_cycle < kubo_afraid){
		flg = false;
		rcsc::dlog.addText( rcsc::Logger::DRIBBLE,
								  "%s:%d: !afraid diff= %d, afraid= %d!"
								  ,__FILE__, __LINE__,
								  opp_cycle-self_cycle, kubo_afraid);
	 }
	 if(flg) break;
  }

  return flg;


    return true;
}
