#include "body_safety_pass.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/math_util.h>

#include "kubo_interception.h"
#include <rcsc/player/say_message_builder.h>

#include <iostream>
using namespace std;

bool
Body_SafetyPass::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_SafetyPass"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    const rcsc::PlayerPtrCont & teammates = wm.teammatesFromSelf();
    const rcsc::PlayerPtrCont::const_iterator teammates_end = teammates.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = teammates.begin();
          it != teammates_end;
          ++it )
    {
        double speed;
        if( checkPass(agent,*it) ){
            if( !decideKickSpeed(agent,*it,1.5,speed) ) //magic number
                return false;

            //only one step kick
            if( M_one_step )
            {
                if( rcsc::Body_KickOneStep( (*it)->pos(),speed ).execute(agent) )
                {
                    if( (*it)->unum() >= 0 && (*it)->unum() <= 11 )
                    {
                        agent->addSayMessage( new rcsc::PassMessage( (*it)->unum(),
                                                                     (*it)->pos(),
                                                                     agent->effector().queuedNextBallPos(),
                                                                     agent->effector().queuedNextBallVel() ) );
                    }
                
                    return true;
                }
                else
                {
                    continue;
                }
            }

            //multi step kick
            if( rcsc::Body_KickMultiStep( (*it)->pos(),speed ).execute(agent) ){

                if( (*it)->unum() >= 0 && (*it)->unum() <= 11 )
                {
                    rcsc::dlog.addText( rcsc::Logger::SHOOT,
                                        "%s:%d: pass to %d(%f,%f)"
                                        ,__FILE__, __LINE__,
                                        (*it)->unum(),
                                        (*it)->pos().x,
                                        (*it)->pos().y);

                    agent->addSayMessage( new rcsc::PassMessage( (*it)->unum(),
                                                                 (*it)->pos(),
                                                                 agent->effector().queuedNextBallPos(),
                                                                 agent->effector().queuedNextBallVel() ) );
                }
                
                return true;
            } else {
                continue;
            }
        }


    }
    
    return false;
}

//player˰˥ѥФ뤫å
bool
Body_SafetyPass::checkPass( rcsc::PlayerAgent * agent,
                            const rcsc::PlayerObject * player )
{
    const rcsc::WorldModel & wm = agent->world();

    if( player->posCount() >= 3 ) return false;

    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
    for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
          it != opps_end;
          ++it )
    {
        
        if( !rcsc::Circle2D( (player->pos() + wm.self().pos()) / 2.0,
                             player->pos().dist(wm.self().pos()) / 2.0 + 1.0 
                ).contains( (*it)->pos() ) )
        {
            continue;
        } else {
            return false;
        }

    }

    return true;
}

bool
Body_SafetyPass::decideKickSpeed( rcsc::PlayerAgent * agent,
                                  const rcsc::PlayerObject * player,
                                  const double & target_speed,
                                  double & speed)
{
    speed = target_speed + player->pos().dist(agent->world().self().pos())
        * ( 1 - rcsc::ServerParam::i().ballDecay() );

    rcsc::Vector2D tmp = rcsc::Body_KickOneStep::get_max_possible_vel( (player->pos() - agent->world().self().pos()).th(),
                                                               agent->world().self().kickRate(),
                                                               agent->world().ball().vel() );

    if( M_one_step && speed > tmp.r() ) return false; 
    
    if( speed <= 0 || speed > rcsc::ServerParam::i().ballSpeedMax() )
        return false;
    else
        return true;
    
}
