#include "body_zulu.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/math_util.h>
#include <rcsc/action/body_kick_multi_step.h>

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_ball.h>

#include "offense_system.h"
#include "kubo_interception.h"
#include <rcsc/player/free_message.h>
#include <sstream>

#include <iostream>
using namespace std;


bool
Body_Zulu::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    //set zulu point
    if(wm.self().pos().x >= -10 && wm.self().pos().x >= wm.offsideLineX() - 20
        && wm.self().pos().x <= 30)
    {
        setZuluPoint(agent,5,5,20);
        setZuluPoint(agent,10,3,20);
        setZuluPoint(agent,15,3,20);
    }
    else if(wm.self().pos().x >= -10
            && wm.self().pos().x >= wm.offsideLineX() - 20
            &&wm.self().pos().x <= 47)
    {
        setZuluPoint(agent,2,5,10);
        setZuluPoint(agent,3,3,12);
        setZuluPoint(agent,4,3,12);
    }
    
    if( zulu_info.size() == 0 ) return false;
    
    rcsc::dlog.addText( rcsc::Logger::PASS,
                        "%s:%d: Bhv_Zulu"
                        ,__FILE__, __LINE__ );
    

    //prepare
    if( !predictTeammates(agent,5) ){     
        return false;
    }
    if( !decideKickSpeed(agent) ){    
        return false;
    }
    if( !predictOpponents(agent,10) ){
        return false;
    }
    decideZuluPoint();

    //execute
    if(zulu_idx != -1
       && zulu_info[zulu_idx].receiver_pos.x < wm.offsideLineX()+1)
    {
//         cout << "[" << wm.self().unum() << "," << wm.time().cycle()
//              << "] " << "zulu" << endl;
        agent->debugClient().addMessage( "ZULU%d",
                                         zulu_info[zulu_idx].receiver_unum );
        agent->debugClient().addCircle( zulu_info[zulu_idx].zulu_point,2 );
        
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: !! zulu !!"
                            ,__FILE__, __LINE__ );
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: receiver_unum %d, opp_unum %d"
                            ,__FILE__, __LINE__,
                            zulu_info[zulu_idx].receiver_unum,
                            zulu_info[zulu_idx].opp_unum);
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: zulu_point (%f,%f)"
                            ,__FILE__, __LINE__,
                            zulu_info[zulu_idx].zulu_point.x,
                            zulu_info[zulu_idx].zulu_point.y);
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: cycle_margin %f"
                            ,__FILE__, __LINE__, cycle_margin );

        //say
        agent->addSayMessage( new rcsc::FreeMessage<8>(decideSendMessage()) );

       
        return rcsc::Body_KickMultiStep(zulu_info[zulu_idx].zulu_point,
                                        std::min(zulu_info[zulu_idx].first_speed,rcsc::ServerParam::i().ballSpeedMax()) , //magic number
                                        true).execute(agent);
    } else {
        return false;
    }
    
}

//offsidelinepass_line
//pass_width,ֳpass_intervalä
void
Body_Zulu::setZuluPoint(rcsc::PlayerAgent * agent,
                        const double & pass_line,
                        const double & pass_interval,
                        const double & pass_width)
{
    const rcsc::WorldModel & wm = agent->world();
    rcsc::Vector2D pass_point;
    double tmp = 0;

    if(pass_width <= 0 || pass_interval <= 0 || pass_width > 30) return;
    
    pass_point.x = wm.offsideLineX() + pass_line;
    tmp = -pass_width;

    if( pass_point.x < -35 || pass_point.x > 52 ) return;
    
    while(1)
    {
        pass_point.y = tmp;
        
        //zuluPoint̵̤äݤȤϤǾʤ(magic number)
        if( pass_point.x > 35
            && abs(pass_point.y) > pass_point.x * (-2) + 104 ){
            tmp += pass_interval;
            if(tmp > pass_width) break;
            continue;
        }
        if( pass_point.x > 47 && abs(pass_point.y) > 10){
            tmp += pass_interval;
            if(tmp > pass_width) break;
            continue;
        }
        //---------------------------------------------------//
    
        zulu_info.push_back( ZuluInfo(pass_point) );
        
        agent->debugClient().addCircle( pass_point,1 );
        
        tmp += pass_interval;
        if(tmp > pass_width) break;
    }
    
}

//zuluݥȤ˺®ãǤ̣Ĵ٤
bool
Body_Zulu::predictTeammates(rcsc::PlayerAgent * agent,
                            const int & count_thr)
{
    
    for(vector< ZuluInfo >::iterator it = zulu_info.begin();
        it != zulu_info.end();++it)
    {
        //skip disable point
        if( !(*it).valid ) continue;
        
        if( !OffenseSystem::pointPredict(agent,(*it).zulu_point,count_thr,TEAM,
                                  &((*it).receiver_cycle),
                                  &((*it).receiver_unum)) ){
            (*it).valid = false;
            continue;
        }
        
        if(  agent->world().teammate((*it).receiver_unum) ==
             static_cast< rcsc::PlayerObject * >( 0 ) )
            continue;
        
        (*it).receiver_pos =
            agent->world().teammate((*it).receiver_unum)->pos();
    }

    return true;
}


//ܡν®
bool
Body_Zulu::decideKickSpeed(rcsc::PlayerAgent * agent)
{
    const rcsc::WorldModel & wm = agent->world();
    double distance=0;
    double min_speed;
    rcsc::Vector2D baller_pos=wm.self().pos();
    double ball_step=0;

   for(vector< ZuluInfo >::iterator it = zulu_info.begin();
        it != zulu_info.end();++it)
    {
        //skip disable point
        if( !(*it).valid ) continue;

        distance = (baller_pos-(*it).zulu_point).r();  //dist from self to point
        min_speed = ((1-rcsc::ServerParam::i().ballDecay())/
                     (1-pow(rcsc::ServerParam::i().ballDecay(),
                            (*it).receiver_cycle+1)))*distance;
	    
        if(min_speed <= rcsc::ServerParam::i().ballSpeedMax()){
            ball_step = rcsc::calc_length_geom_series( min_speed,
                                                       distance,
                                                       rcsc::ServerParam::i().ballDecay());
            (*it).ball_cycle = ball_step;
            (*it).first_speed = min_speed;
        }
        else
        {
            (*it).valid = false;
            continue;
        }
        
    }
   
   return true;
}


//zuluݥ
bool
Body_Zulu::predictOpponents(rcsc::PlayerAgent * agent,
                            const int & count_thr){
    
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerCont & opponents = wm.opponents();
    const rcsc::PlayerType * player_type;
    double fast_cycle, kickable_area;
    int fast_unum=-1;
    bool flg;

    
    for(vector< ZuluInfo >::iterator itz = zulu_info.begin();
        itz != zulu_info.end();++itz)
    {
        //skip disable point
        if( !(*itz).valid ) continue;
    
        fast_cycle = 1000;
        flg = false;
        
        for ( rcsc::PlayerCont::const_iterator it = opponents.begin();
              it != opponents.end();
              ++it )
        {
            rcsc::dlog.addText( rcsc::Logger::PASS,
                                "%s:%d: !!predictOpp unum%d(%f,%f) count%d!!"
                                ,__FILE__, __LINE__, (*it).unum(),
                                (*it).pos().x,(*it).pos().y,(*it).posCount());
            //check unum
            if ( (*it).unum() == -1 ){
                continue;
            }

            //check posCount
            if ( (*it).posCount() > count_thr){
                continue;
            }
            
            player_type = (*it).playerTypePtr();
            
            //check player_type
            if ( ! player_type ){
                continue;
            }
            
            //check goalie
            if((*it).goalie()){
                kickable_area = rcsc::ServerParam::i().catchableArea();
            } else {
                kickable_area = player_type->kickableArea();
            }
            
            //calc kick step
            rcsc::Vector2D result_pos,result_vel;
            int kick_step;
            OffenseSystem::simulateKickMultiStep(agent,
                                          (*itz).zulu_point,
                                          (*itz).first_speed,
                                          result_pos,
                                          result_vel,
                                          kick_step );
            double accel = 100 * player_type->dashPowerRate()
                * player_type->effortMax();
            
            //calc intercept cycle
            KuboInterception opp_inter(result_pos,result_vel);
            double cycle = opp_inter.getReachCycleKubo( (*it).pos(),
                                                        &((*it).vel()),
                                                        &((*it).body()),
                                                        player_type->playerDecay(),
                                                        player_type->inertiaMoment(),
                                                        accel,
                                                        (*it).posCount(),
                                                        kickable_area,
                                                        player_type->realSpeedMax());
            rcsc::dlog.addText( rcsc::Logger::PASS,
                                "%s:%d: !!cycle=%f!!"
                                ,__FILE__, __LINE__, cycle);
            if(cycle < fast_cycle){
                fast_cycle = cycle;
                fast_unum = (*it).unum();
                flg = true;
            }
        }

        if(flg){
            (*itz).opp_cycle = fast_cycle;
            (*itz).opp_unum = fast_unum;
            if(  wm.teammate((*itz).opp_unum) ==
                 static_cast< rcsc::PlayerObject * >( 0 ) ){
                (*itz).opp_pos = wm.opponent((*itz).opp_unum)->pos();
            } else {
                (*itz).valid = false;
                continue;
            }
        } else {
            (*itz).valid = false;
            continue;
        }
        
    }
    
    return true;
}
        
//ball_cycleopp_cycleκ礭zulu_pointõ
void
Body_Zulu::decideZuluPoint()
{
    double tmp;
    
    for(unsigned int i=0;i<zulu_info.size();i++)
    {
        //skip disable point
        if( !zulu_info[i].valid ) continue;
        
        if(zulu_info[i].zulu_point.x <= 47){
            tmp = zulu_info[i].opp_cycle - zulu_info[i].ball_cycle;

            if(cycle_margin < tmp){
                cycle_margin = tmp;
                zulu_idx = i;
            }
        }
    }

}

string
Body_Zulu::decideSendMessage()
{
    int px=0;
    int py=0;
    string send_message="";
    string renum="";

    ////////////////send zulupoint decide/////////////////////
    //ͼθ
    if(zulu_info[zulu_idx].zulu_point.x > 0){
        px=(int)(zulu_info[zulu_idx].zulu_point.x+0.5);
    } else {
        px=(int)(zulu_info[zulu_idx].zulu_point.x-0.5);
    }

    if(zulu_info[zulu_idx].zulu_point.y>0){
        py=(int)(zulu_info[zulu_idx].zulu_point.y+0.5);
    } else {
        py=(int)(zulu_info[zulu_idx].zulu_point.y-0.5);
    }
    //////////////////////////////////////////////////////////

    px += 50;
    py += 50;

    if( px >= 100 || py >= 100 )return false;
    
    std::ostringstream os;

    if(abs(px) < 10)
        os << "0" << px;
    else
        os << px;
    
    if(abs(py) < 10)
        os << "0" << py;
    else
        os << py;

    if(zulu_info[zulu_idx].receiver_unum < 10)
        os << "0" << zulu_info[zulu_idx].receiver_unum;
    else
        os << zulu_info[zulu_idx].receiver_unum;
    
    if(zulu_info[zulu_idx].receiver_cycle < 10)
        os << "0" << zulu_info[zulu_idx].receiver_cycle;
    else
        os << zulu_info[zulu_idx].receiver_cycle;

    send_message = os.str();
    if(send_message.size() != 8) send_message="oooooooo";

    return send_message;
}


//zuluιޤʹư
bool
Body_Zulu::hear( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();
    static bool S_offense_flg = false;
    static int S_start_cycle = 0;
    static bool S_intercept_flg = false;
    static rcsc::Vector2D S_dash_point;
    static int S_dash_cycle = 0;

    if( wm.time().cycle() == 0 ) return false;

    if(wm.time().cycle() == 0) return false;
    
    //attention to kickable teammatex
    if( !wm.teammatesFromBall().empty() 
        && (*(wm.teammatesFromBall().begin()))->unum() >= 1
        && (*(wm.teammatesFromBall().begin()))->unum() <= 11 )
    {
        agent->doAttentionto(rcsc::LEFT,
                             (*(wm.teammatesFromBall().begin()))->unum());
    }
        
    if ( wm.audioMemory().freeMessageTime() == agent->world().time()
         && wm.audioMemory().freeMessage().front().message_.size() == 8)
    {
        string message = wm.audioMemory().freeMessage().front().message_;
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: !!receive move_message=%s!!"
                            ,__FILE__, __LINE__, message.c_str() );
        if( message == "oooooooo" ) return false;

        int receiver_unum = atoi(read(message,4,5).c_str());

        if(receiver_unum == wm.self().unum()){
            S_dash_point.x = atoi(read(message,0,1).c_str()) - 50;
            S_dash_point.y = atoi(read(message,2,3).c_str()) - 50;
            S_dash_cycle = atoi(read(message,6,7).c_str());
            S_offense_flg = true;
            S_start_cycle = wm.time().cycle();
            S_intercept_flg = false;
        }
    }
        

    int cycle_after_hear = wm.time().cycle() - S_start_cycle;
    
    //stop move
    if(S_offense_flg && cycle_after_hear >= S_dash_cycle/2) //magic number
    {
        S_offense_flg = false;
        if(cycle_after_hear == S_dash_cycle/2){
            S_intercept_flg = true;
        }
    }

    //stop intercept
    if(S_intercept_flg && cycle_after_hear > S_dash_cycle + 3){ //magic number
        if( wm.existKickableOpponent()
            || wm.self().pos().dist( wm.ball().pos() ) > 4 ){
            S_intercept_flg = false;
        }
    }

    if(S_intercept_flg)
    {
        if( wm.existKickableOpponent() ){
            S_offense_flg = false;
            S_intercept_flg = false;
            return false;
        }
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: !!intercept cycle = %d!!"
                            ,__FILE__, __LINE__, cycle_after_hear);
        rcsc::Body_Intercept().execute( agent );
        return true;
    }

    if(S_offense_flg)
    {
        if( wm.existKickableOpponent() ){
            S_offense_flg = false;
            S_intercept_flg = false;
            return false;
        }
        
        double dist_thr = wm.ball().distFromSelf() * 0.1;
        if ( dist_thr < 1.0 ) dist_thr = 1.0;
        
        rcsc::dlog.addText( rcsc::Logger::PASS,
                            "%s:%d: !!moving to (%f,%f) cycle = %d!!"
                            ,__FILE__, __LINE__, S_dash_point.x, S_dash_point.y,
                            cycle_after_hear);
        
        if ( ! rcsc::Body_GoToPoint(S_dash_point, dist_thr,
                                    rcsc::ServerParam::i().maxPower()
                 ).execute( agent ) )
        {
            rcsc::Body_TurnToBall().execute( agent );
        }
        return true;
    }

    return false;
}


string
Body_Zulu::read(string str,int first,int last)
{
    string str_tmp = "";
    for(int i=first;i<=last;i++)
    {
        str_tmp += str[i];
    }

    return str_tmp;
}

