#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "kubo_interception.h"

#include <rcsc/common/server_param.h>
#include <rcsc/common/player_type.h>
#include <rcsc/common/logger.h>
#include <rcsc/soccer_math.h>
#include <rcsc/math_util.h>

#include <algorithm>
#include <iostream>
#include <cmath>

//#define TRACE_NEWTON

//namespace rcsc {

    const double KuboInterception::NEVER_CYCLE = 1000.0;


    namespace {

/*-------------------------------------------------------------------*/
/*!
  \brief get the value of log(ball_decay)
  accessible only from this file.
*/
        inline
        const
        double & logBallDecay()
        {
            static const double LOG_BALL_DECAY = std::log( rcsc::ServerParam::i().ballDecay() );
            return LOG_BALL_DECAY;
        }

    };

/*-------------------------------------------------------------------*/
/*!

*/
    KuboInterception::KuboInterception( const rcsc::Vector2D & ball_pos,
                                        const double & ball_first_speed,
                                        const rcsc::AngleDeg & ball_vel_angle )
        : M_ball_first_pos( ball_pos )
        , M_ball_first_speed( ball_first_speed )
        , M_ball_vel_angle( ball_vel_angle )
        , M_ball_x_constant( ball_first_speed
                             / (1.0 - rcsc::ServerParam::i().ballDecay()) )
        , M_ball_x_d_constant( (-ball_first_speed * logBallDecay())
                               / (1.0 - rcsc::ServerParam::i().ballDecay()) )
    {

    }

/*-------------------------------------------------------------------*/
/*!

*/
    KuboInterception::KuboInterception( const rcsc::Vector2D & ball_pos,
                                        const rcsc::Vector2D & ball_vel )
        : M_ball_first_pos( ball_pos )
        , M_ball_first_speed( ball_vel.r() )
        , M_ball_vel_angle( ball_vel.th() )
        , M_ball_x_constant( M_ball_first_speed
                             / (1.0 - rcsc::ServerParam::i().ballDecay()) )
        , M_ball_x_d_constant( (-M_ball_first_speed * logBallDecay())
                               / (1.0 - rcsc::ServerParam::i().ballDecay()) )
    {

    }

/*-------------------------------------------------------------------*/
/*!

*/
//     double
//     KuboInterception::getReachCycle( const rcsc::Vector2D & player_pos,
//                                      const rcsc::Vector2D * player_vel,
//                                      const rcsc::AngleDeg * player_angle,
//                                      const int player_count,
//                                      const double & control_buf,
//                                      const double & player_max_speed ) const
//     {
//         static const int MAX_LOOP = 20;
//         static const double MIN_ERROR = 0.05;

//         // player related param
//         /*
//           const double player_max_speed
//           = ( player_type
//           ? player_type->playerSpeedMax()
//           : rcsc::ServerParam::i().defaultPlayerSpeedMax() );
//           const double control_buf
//           = ( goalie
//           ? rcsc::ServerParam::i().catchAreaLength()
//           : ( player_type
//           ? ( rcsc::ServerParam::i().ballSize()
//           + player_type->playerSize()
//           + player_type->kickableMargin() )
//           : rcsc::ServerParam::i().defaultKickableArea() ) );
//         */

//         // relative to ball pos and ball velocity angle
//         rcsc::Vector2D start_point = player_pos;
//         start_point -= M_ball_first_pos;

//         ///////////////////////////////////////////
//         // check kickable state at current
//         if ( start_point.r2() < rcsc::square( control_buf ) )
//         {
// #ifdef TRACE_NEWTON
//             cout << "  Newton: ZERO" << endl;
// #endif
//             return 0.0;
//         }


//         // rotation
//         start_point.rotate( - M_ball_vel_angle );


//         // consider player confidence
//         if ( player_count == 0 )
//         {
//             // if player is almost all stopped,
//             // it is necessary for him to accelarate his dash speed at first.
//             if ( player_vel && player_vel->r() < 0.1 )
//             {
//                 if ( start_point.absY() > control_buf )
//                 {
//                     start_point.y += player_max_speed * ( start_point.y > 0.0 ? 1.0 : -1.0 );
//                 }
//                 else
//                 {
//                     start_point.x += player_max_speed * ( start_point.x > 0.0 ? 1.0 : -1.0 );
//                 }
//             }
//         }
//         else
//         {
//             double virtual_dash = player_max_speed * 0.8;
//             virtual_dash *= std::min(5, player_count);
//             if ( start_point.absY() > virtual_dash )
//             {
//                 start_point.y -= virtual_dash * ( start_point.y > 0.0 ? 1.0 : -1.0 );
//             }
//             else
//             {
//                 double x_dash = std::sqrt( rcsc::square( virtual_dash )
//                                            - rcsc::square( start_point.y ) );
//                 start_point.y = 0.0;
//                 if ( start_point.absX() > x_dash )
//                 {
//                     start_point.x -= x_dash * ( start_point.x > 0.0 ? 1.0 : -1.0 );
//                 }
//                 else
//                 {
//                     start_point.x = 0.0;
//                 }
//             }
//         }



//         /////////////////////////////////////////////////////////////////////////
//         // Newton method

//         const double start_y2 = rcsc::square( start_point.y );

//         const double bdecay = rcsc::ServerParam::i().ballDecay();

//         // t: cycles to reach the ball
//         // f: distance(ball_pos, player_first_pos) - (player_total_dash)
//         // f_d: differential of f

//         double t = 0.0;
//         double f, f_d;
//         int counter = 0;
//         do
//         {
//             counter++;
//             //double ball_x
//             //    = util::inertia_n_step_distance( M_ball_first_speed, t, bdecay );
//             double ball_x = M_ball_x_constant * ( 1.0 - std::pow( bdecay, t) );
//             //double ball_x_d
//             //    = (-M_ball_first_speed * std::pow( bdecay, t ) * LOG_BALL_DECAY )
//             //    / (1.0 - rcsc::ServerParam::i().ballDecay() );
//             double ball_x_d = M_ball_x_d_constant * std::pow( rcsc::ServerParam::i().ballDecay(), t );
//             // distance to ball at t from player first pos
//             double dist_to_ball = std::sqrt( rcsc::square( ball_x - start_point.x )
//                                              + start_y2 );
//             f = dist_to_ball - player_max_speed * t - control_buf;
//             f_d = (ball_x - start_point.x) * ball_x_d / dist_to_ball - player_max_speed;
//             if ( ball_x < start_point.x && f_d != 0.0 )
//             {
//                 //t = t - f / f_d;
//                 t -= f / f_d;
// #ifdef TRACE_NEWTON
//                 cout << "  Newton:" << counter << ": update_1 cyc=" << t
//                      << "  f=" << f
//                      << "  f_d=" << f_d
//                      << endl;
// #endif
//             }
//             else if ( ball_x > start_point.x && f_d < 0.0 )
//             {
//                 t -= f / f_d;
// #ifdef TRACE_NEWTON
//                 cout << "  Newton:" << counter << ":  update_2 cyc=" << t
//                      << "  f=" << f
//                      << "  f_d=" << f_d
//                      << endl;
// #endif
//             }
//             else
//             {
//                 t += f / player_max_speed;
// #ifdef TRACE_NEWTON
//                 cout << "  Newton:" << counter << ":  update_3 cyc=" << t
//                      << "  f=" << f
//                      << "  f_d=" << f_d;
// #endif
//                 // if differential of f is positive value,
//                 // ball is moving far away from player faster than player's speed.
//                 // And if f_d is small, Newton method required many loop.
//                 // so, it is good idea to add the cycles enforcely.
//                 if ( f_d > 0.0
//                      // || ( f > 0.01 && fabs(f_d) < 0.01 )
//                     )
//                 {
//                     t += 10.0; // XXX Magic Number XXX
// #ifdef TRACE_NEWTON
//                     cout << "  update enforcely. cyc=" << t;
// #endif
//                 }
// #ifdef TRACE_NEWTON
//                 cout << endl;
// #endif
//             }

//             if ( fabs(f) < MIN_ERROR )
//             {
//                 break;
//             }
//         }
//         while ( counter < MAX_LOOP );


//         if ( std::fabs(f) > MIN_ERROR )
//         {
//             return NEVER_CYCLE;
//         }

//         return t;
//     }

/*-------------------------------------------------------------------*/
/*!
  testing purpose
*/
//     double
//     KuboInterception::getReachCycleGreedly( const rcsc::Vector2D & player_pos,
//                                             const rcsc::Vector2D * player_vel,
//                                             const rcsc::AngleDeg * player_angle,
//                                             const int player_count,
//                                             const double & control_buf,
//                                             const double & player_max_speed ) const
//     {
//         // player related param
//         /*
//           const double player_max_speed
//           = ( player_type
//           ? player_type->playerSpeedMax()
//           : rcsc::ServerParam::i().defaultPlayerSpeedMax() );
//           const double control_buf
//           = ( goalie
//           ? rcsc::ServerParam::i().catchAreaLength()
//           : ( player_type
//           ? ( rcsc::ServerParam::i().ballSize()
//           + player_type->playerSize()
//           + player_type->kickableMargin() )
//           : rcsc::ServerParam::i().defaultKickableArea() ) ) - 0.05;
//         */

//         rcsc::Vector2D ball_pos( M_ball_first_pos );
//         rcsc::Vector2D ball_vel
//             = rcsc::Vector2D::polar2vector( M_ball_first_speed, M_ball_vel_angle );

//         rcsc::Vector2D start_point = player_pos;
//         start_point -= M_ball_first_pos;
//         start_point.rotate( -M_ball_vel_angle );
//         const double player_cyc_to_intersect
//             = std::max(0.0,
//                        std::ceil( ( start_point.absY() - control_buf * 0.5 )
//                                   / player_max_speed )
//                 );
//         double cycle = std::max( 0.0, player_cyc_to_intersect - 1.0 );
//         double player_dash = player_max_speed * cycle;
//         ball_pos = inertia_n_step_point( ball_pos,
//                                          ball_vel,
//                                          (int)cycle,
//                                          rcsc::ServerParam::i().ballDecay() );
//         ball_vel *= std::pow( rcsc::ServerParam::i().ballDecay(), cycle );
// #ifdef TRACE_NEWTON
//         cout << "  Greedy: start=" << cycle << endl;
// #endif
//         while ( cycle < 100.0 )
//         {
//             if ( player_pos.dist2( ball_pos ) < rcsc::square( player_dash + control_buf ) )
//             {
// #ifdef TRACE_NEWTON
//                 cout << "  Greedy: " << cycle
//                      << "  dist_to_ball: " << player_pos.dist(ball_pos)
//                      << "  player_dash+ctrl: " << player_dash + control_buf
//                      << endl;
// #endif
//                 return cycle;
//             }

//             player_dash += player_max_speed;
//             ball_pos += ball_vel;
//             ball_vel *= rcsc::ServerParam::i().ballDecay();
//             cycle += 1.0;
//         }


//         return NEVER_CYCLE;
//     }

//}




////////////////added by okubo/////////////

double
KuboInterception::getReachCycleKubo( const rcsc::Vector2D & player_pos,
                                     const rcsc::Vector2D * player_vel,
                                     const rcsc::AngleDeg * player_angle,
                                     const double player_decay,
                                     const double inertia_moment,
                                     const double dash_max_power,
                                     const int player_count,
                                     const double & control_buf,
                                     const double & player_max_speed) const
{
    static const int MAX_LOOP = 20;
    static const double MIN_ERROR = 0.05;
    // relative to ball pos and ball velocity angle
    rcsc::Vector2D start_point = player_pos;
    start_point -= M_ball_first_pos;

    ///////////////////////////////////////////
    // check kickable state at current
    if ( start_point.r2() < rcsc::square( control_buf ) )
    {
#ifdef TRACE_NEWTON
        cout << "  Newton: ZERO" << endl;
#endif
        return 0.0;
    }

  

    /////////////////////////////////////////////////////////////////////////
    // Newton method

    const double b_decay = rcsc::ServerParam::i().ballDecay();

    // t: cycles to reach the ball
    // f: distance(ball_pos, player_first_pos) - (player_total_dash)
    // f_d: differential of f

    double t = 0.0;
    double f, f_d;
    int counter = 0;
    rcsc::Vector2D v_now;
    double p_decay=player_decay;
  
    do
    {
        start_point = player_pos-M_ball_first_pos;
        counter++;
        //double ball_x
        //    = util::inertia_n_step_distance( M_ball_first_speed, t, bdecay );
        double ball_x = M_ball_x_constant * ( 1.0 - std::pow( b_decay, t) );
        //double ball_x_d
        //    = (-M_ball_first_speed * std::pow( bdecay, t ) * LOG_BALL_DECAY )
        //    / (1.0 - rcsc::ServerParam::i().ballDecay() );
        double ball_x_d = M_ball_x_d_constant * std::pow(b_decay, t );
        // distance to ball at t from player first pos
        //double dist_to_ball = std::sqrt( rcsc::square( ball_x - start_point.x )
        //                               + start_y2 );

        int turncount=0;
    
        /////////startpoint change/////////
        if(player_vel){
            v_now = *player_vel;
        } else {
            return 1000;
        }
        double movedist=( (player_vel->r()*(1-pow(p_decay,t))) / (1-p_decay) );
        start_point=start_point+v_now.setLength(movedist);
        
        //////////turncount decide//////////////////
        rcsc::Vector2D intercept_point;
        intercept_point.setPolar(M_ball_first_speed,M_ball_vel_angle);
        intercept_point.setLength(ball_x);  
        intercept_point=intercept_point-start_point;
        rcsc::AngleDeg target_angle=intercept_point.th();
        double actual_angle=fabs(180/(1.0+inertia_moment* player_vel->r()));
        double ang=fabs((target_angle-*player_angle).degree());
      
        if(ang>180)ang=360-ang;
      
        int count=0;
        bool turn_flg = false;
        while(1)
        {
            //check bodyCount
            if( player_count >= 3 ){ //magic number
                turn_flg = true;
                turncount = 1;
                break;
            }
            
            actual_angle = fabs( 180 / (1.0 + inertia_moment
                                        *player_vel->r() *pow(p_decay,count)));
            count++;
            ang -= actual_angle;

            if(ang < 0){  //finish turn
                turn_flg = true;
                turncount = count;
                break;
            } 
            else if(count > 5){
                turn_flg = false;
                break;
            }
        }

        //turn failed
        if(!turn_flg){    
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: turn_flg is false"
                                ,__FILE__, __LINE__ );
            return 1000;
        }
     

        ///////////roatation/////////////////////////////////////
        rcsc::Vector2D start_point1=start_point;
     
        start_point.rotate( - M_ball_vel_angle );
	
       
        double start_y2 = rcsc::square( start_point.y );
        double dist_to_ball = std::sqrt( rcsc::square( ball_x - start_point.x )
                                         + start_y2);

        ///////////////move distance//////////

        double realt=t-turncount;
        double formula=0;
        double formula_d=0;
     
        if(t>turncount)
        {
            formula=(dash_max_power/(1-p_decay))*
                (realt-(p_decay*(1-pow(p_decay,realt)))/(1-p_decay));
            formula_d=(dash_max_power/(1-p_decay))*
                (1+((std::log(p_decay)/(1-p_decay))*pow(p_decay,realt+1)));
        }
        else
        {
            formula=0;
            formula_d=0;
        }

     

        //////////////////////////////////////
        //f = dist_to_ball - movedistance - control_buf;
        f = dist_to_ball - formula - control_buf;
        f_d =(ball_x-start_point.x)*ball_x_d/dist_to_ball-formula_d;
    
        //f = dist_to_ball - player_max_speed * t - control_buf;
        //f_d = (ball_x - start_point.x) * ball_x_d / dist_to_ball - player_max_speed;
      

        //rcsc::Vector2D player_intercept_point = intercept_point;
        //player_intercept_p = player_intercept_p.setLength(formula)
        //    + start_point1 + M_ball_first_pos;
    
        if ( fabs(f) < MIN_ERROR )
        {
            break;
        }


        if ( ball_x < start_point.x && f_d != 0.0 )
        {
            //t = t - f / f_d;
            t -= f / f_d;
#ifdef TRACE_NEWTON
            cout << "  Newton:" << counter << ": update_1 cyc=" << t
                 << "  f=" << f
                 << "  f_d=" << f_d
                 << endl;
#endif
        }
        else if ( ball_x > start_point.x && f_d < 0.0 )
        {
            t -= f / f_d;
#ifdef TRACE_NEWTON
            cout << "  Newton:" << counter << ":  update_2 cyc=" << t
                 << "  f=" << f
                 << "  f_d=" << f_d
                 << endl;
#endif
        }
        else
        {
            t += f / player_max_speed;
            //t += f /formula_d;
#ifdef TRACE_NEWTON
            cout << "  Newton:" << counter << ":  update_3 cyc=" << t
                 << "  f=" << f
                 << "  f_d=" << f_d;
#endif
            // if differential of f is positive value,
            // ball is moving far away from player faster than player's speed.
            // And if f_d is small, Newton method required many loop.
            // so, it is good idea to add the cycles enforcely.
            if ( f_d > 0.0
                 // || ( f > 0.01 && fabs(f_d) < 0.01 )
                )
            {
                t += 10.0; // XXX Magic Number XXX
#ifdef TRACE_NEWTON
                cout << "  update enforcely. cyc=" << t;
#endif
            }
#ifdef TRACE_NEWTON
            cout << endl;
#endif
        }
      
    }
    while ( counter < MAX_LOOP );


    if ( std::fabs(f) > MIN_ERROR  || t<0)
    {
        return 1000;
    }

 
    return t;
}




///////////////////////////////////////

