// -*-c++-*-

/*!
  \file player_change_state.h
  \brief player change state holder class Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_COACH_PLAYER_CHANGE_STATE_H
#define RCSC_COACH_PLAYER_CHANGE_STATE_H

#include <map>

namespace rcsc {

class GameTime;
class GameMode;

/*!
  \class PlayerChangeState
  \brief the holder of player change status
 */
class PlayerChangeState {
private:
    int M_change_count;

    // key: type id
    // value: used count of that type
    std::map< int, int > M_teammate_type_count;

    int M_teammate_types[11];
    bool M_opponent_changed[11];


    // nocopyable
    PlayerChangeState( const PlayerChangeState & );
    PlayerChangeState & operator=( const PlayerChangeState & );
public:
    PlayerChangeState();

    // setter methods

    // parse (ok change_player_type ...) or (change_player_type ...
    void parse( const char * msg,
                const GameMode & game_mode,
                const GameTime & current );
private:
    void change(const int unum, const int type);

public:
    // getter methos

    bool canChange( const int unum,
                    const int type,
                    const GameMode & game_mode,
                    const GameTime & current ) const;

    int getTeammateType( const int unum ) const;
    bool isOpponentChanged( const int unum ) const;


};

}

#endif
